import { constants, helpers } from 'ptokens-utils';
import { Node } from 'ptokens-node';
import { HttpProvider } from 'ptokens-providers';

const BOOT_TESTNET_ENDPOINT = 'https://testnet--bootnode-eu-1.p.network';
const BOOT_MAINNET_ENDPOINT = 'https://mainnet--bootnode-eu-1.p.network';
const getBootNodeEndpoint = _network => {
  return _network === constants.networks.Mainnet ? BOOT_MAINNET_ENDPOINT : BOOT_TESTNET_ENDPOINT;
};

class NodeSelector {
  constructor(_configs) {
    if (_configs) this.setParams(_configs);
    this.nodes = [];
    this.provider = new HttpProvider();
  }
  async checkConnection(_endpoint, _timeout, _options = {}) {
    const {
      pToken: optionalPtoken,
      hostBlockchain: optionalHostBlockchain,
      hostNetwork: optionalHostNetwork,
      nativeBlockchain: optionalNativeBlockchain,
      nativeNetwork: optionalNativeNetwork
    } = _options;
    try {
      this.provider.setEndpoint(_endpoint);
      const {
        host_blockchain,
        host_network,
        native_blockchain,
        native_network
      } = await this.provider.call('GET', `/${optionalPtoken || this.pToken}-on-${helpers.getBlockchainShortType(optionalHostBlockchain || this.hostBlockchain)}/get-info`, null, _timeout);
      return Boolean(host_blockchain === (optionalHostBlockchain || this.hostBlockchain) && host_network === (optionalHostNetwork || this.hostNetwork) && native_blockchain === (optionalNativeBlockchain || this.nativeBlockchain) && native_network === (optionalNativeNetwork || this.nativeNetwork));
    } catch (_err) {
      throw new Error(`Error during checking node connection: ${_err.message}`);
    }
  }
  async getApi() {
    try {
      if (!this.selectedNode) {
        const node = await this.select();
        if (!node) throw new Error('No Nodes Available');
      }
      return this.selectedNode.api;
    } catch (_err) {
      throw new Error(`Error during getting node api: ${_err.message}`);
    }
  }
  async select(_options = {}) {
    const {
      timeout,
      forceFetchingNodes,
      nodes: optionalNodes,
      pToken: optionalPtoken,
      hostBlockchain: optionalHostBlockchain,
      hostNetwork: optionalHostNetwork,
      nativeBlockchain: optionalNativeBlockchain,
      nativeNetwork: optionalNativeNetwork
    } = _options;
    try {
      if ((this.nodes.length === 0 || forceFetchingNodes) && !optionalNodes) await this.fetchNodes(helpers.getNetworkType(optionalHostNetwork || this.hostNetwork));
      const feature = `${optionalPtoken || this.pToken}-on-${helpers.getBlockchainShortType(optionalHostBlockchain || this.hostBlockchain)}`;
      const filteredNodesByFeature = optionalNodes ? optionalNodes.filter(({
        features
      }) => features.includes(feature)) : this.nodes.filter(({
        features
      }) => features.includes(feature));
      if (filteredNodesByFeature.length === 0) throw new Error(`No nodes available relating to the selected pToken (${feature})`);
      const nodesNotReachable = [];
      for (;;) {
        const index = Math.floor(Math.random() * filteredNodesByFeature.length);
        const selectedNode = filteredNodesByFeature[index];
        if ((await this.checkConnection(selectedNode.webapi, timeout || 30000, {
          pToken: optionalPtoken,
          nativeNetwork: optionalNativeNetwork,
          nativeBlockchain: optionalNativeBlockchain,
          hostNetwork: optionalHostNetwork,
          hostBlockchain: optionalHostBlockchain
        })) && !nodesNotReachable.includes(selectedNode)) {
          return this.setSelectedNode(selectedNode.webapi, {
            pToken: optionalPtoken,
            hostBlockchain: optionalHostBlockchain
          });
        } else if (!nodesNotReachable.includes(selectedNode)) {
          nodesNotReachable.push(selectedNode);
        }
        if (nodesNotReachable.length === filteredNodesByFeature.length) throw new Error(`All nodes relating to the selected pToken (${feature}) appear to be unavailable`);
      }
    } catch (_err) {
      throw new Error(_err.message);
    }
  }
  setSelectedNode(_node, _options = {}) {
    const {
      pToken: optionalPtoken,
      hostBlockchain: optionalHostBlockchain
    } = _options;
    if (_node instanceof Node) {
      this.selectedNode = _node;
      return this.selectedNode;
    }
    this.selectedNode = new Node({
      pToken: optionalPtoken || this.pToken,
      blockchain: optionalHostBlockchain || this.hostBlockchain,
      provider: new HttpProvider(_node)
    });
    return this.selectedNode;
  }
  setParams(_configs) {
    const {
      defaultNode,
      pToken
    } = _configs;
    if (!helpers.isValidPTokenName(pToken)) throw new Error('Invalid pToken name');
    this.pToken = pToken.toLowerCase() === constants.pTokens.pETH ? constants.pTokens.pWETH : pToken.toLowerCase();
    const {
      hostBlockchain,
      hostNetwork,
      nativeBlockchain,
      nativeNetwork
    } = helpers.parseParams(_configs, _configs.nativeBlockchain ? helpers.getBlockchainType[_configs.nativeBlockchain] : helpers.getNativeBlockchainFromPtokenName(this.pToken));
    this.hostBlockchain = hostBlockchain;
    this.hostNetwork = hostNetwork;
    this.nativeBlockchain = nativeBlockchain;
    this.nativeNetwork = nativeNetwork;
    this.selectedNode = defaultNode ? this.setSelectedNode(defaultNode) : null;
  }
  async fetchNodes(_networkType) {
    try {
      this.provider.setEndpoint(getBootNodeEndpoint(_networkType));
      this.nodes = (await this.provider.call('GET', '/peers')).peers;
      return this.nodes;
    } catch (_err) {
      throw new Error(_err.message);
    }
  }
}

export { NodeSelector };
